@@ -21,6 +21,7 @@
 #include <libHX/defs.h>
 #include <libHX/init.h>
 #include <libHX/string.h>
+#include <libHX/io.h>
 #include "config.h"
 #include "cmt-internal.h"
 #include "libcryptmount.h"
@@ -68,9 +69,12 @@ EXPORT_SYMBOL int cryptmount_init(void)
 			return ret;
 		}
 #ifdef HAVE_LIBCRYPTO
+#if OPENSSL_VERSION_NUMBER >= 0x1010000fL
+#else
 		OpenSSL_add_all_algorithms();
 		OpenSSL_add_all_ciphers();
 		OpenSSL_add_all_digests();
+#endif
 #endif
 	}
 	++ehd_use_count;
@@ -106,6 +110,9 @@ EXPORT_SYMBOL int ehd_mtinfo_get(struct ehd_mount_info *mt,
 	case EHD_MTINFO_CRYPTODEV:
 		*static_cast(const char **, ptr) = mt->crypto_device;
 		break;
+	case EHD_MTINFO_HEADER:
+		*static_cast(const char **, ptr) = mt->header;
+		break;
 	case EHD_MTINFO_LOOPDEV:
 		*static_cast(const char **, ptr) = mt->loop_device;
 		break;
@@ -124,6 +131,7 @@ EXPORT_SYMBOL int ehd_mtinfo_get(struct ehd_mount_info *mt,
 EXPORT_SYMBOL void ehd_mtinfo_free(struct ehd_mount_info *mt)
 {
 	free(mt->container);
+	free(mt->header);
 	HXmc_free(mt->crypto_device);
 	HXmc_free(mt->crypto_name);
 	if (mt->loop_device != NULL)
@@ -134,6 +142,28 @@ EXPORT_SYMBOL void ehd_mtinfo_free(struct ehd_mount_info *mt)
 	 */
 }
 
+/**
+ * ehd_realpath - convert path to absolute file name
+ */
+EXPORT_SYMBOL int ehd_realpath(char **path)
+{
+	hxmc_t *tmp = HXmc_strinit("");
+	int ret = HX_realpath(&tmp, *path, HX_REALPATH_DEFAULT | HX_REALPATH_ABSOLUTE);
+	if (ret < 0) {
+		fprintf(stderr, "realpath %s: %s\n", *path, strerror(errno));
+		goto out;
+	}
+	*path = HX_strdup(tmp);
+	if (*path == NULL) {
+		ret = -ENOMEM;
+		goto out;
+	}
+	ret = true;
+ out:
+	HXmc_free(tmp);
+	return ret;
+}
+
 EXPORT_SYMBOL struct ehd_mount_request *ehd_mtreq_new(void)
 {
 	struct ehd_mount_request *rq;
@@ -149,6 +179,7 @@ EXPORT_SYMBOL void ehd_mtreq_free(struct ehd_mount_request *rq)
 {
 	free(rq->container);
 	free(rq->mountpoint);
+	free(rq->header);
 	free(rq->fs_cipher);
 	free(rq->fs_hash);
 	free(rq->key_data);
@@ -170,6 +201,7 @@ EXPORT_SYMBOL int ehd_mtreq_set(struct ehd_mount_request *rq,
 	case EHD_MTREQ_FS_CIPHER:
 	case EHD_MTREQ_FS_HASH:
 	case EHD_MTREQ_FSTYPE:
+	case EHD_MTREQ_HEADER:
 	case EHD_MTREQ_MOUNT_OPTS:
 		orig = va_arg(args, const char *);
 		nv = HX_strdup(orig);
@@ -208,6 +240,12 @@ EXPORT_SYMBOL int ehd_mtreq_set(struct ehd_mount_request *rq,
 	case EHD_MTREQ_ALLOW_DISCARDS:
 		rq->allow_discards = va_arg(args, unsigned int);
 		break;
+	case EHD_MTREQ_NO_READ_WORKQUEUE:
+		rq->no_read_workqueue = va_arg(args, unsigned int);
+		break;
+	case EHD_MTREQ_NO_WRITE_WORKQUEUE:
+		rq->no_write_workqueue = va_arg(args, unsigned int);
+		break;
 	}
 	switch (opt) {
 	case EHD_MTREQ_CONTAINER:
@@ -222,6 +260,10 @@ EXPORT_SYMBOL int ehd_mtreq_set(struct ehd_mount_request *rq,
 		free(rq->container);
 		rq->container = nv;
 		break;
+	case EHD_MTREQ_HEADER:
+		free(rq->header);
+		rq->header = nv;
+		break;
 	case EHD_MTREQ_FS_CIPHER:
 		free(rq->fs_cipher);
 		rq->fs_cipher = nv;
@@ -298,6 +340,11 @@ EXPORT_SYMBOL int ehd_load(struct ehd_mount_request *req,
 		goto out_err;
 	if ((mt->container = HX_strdup(req->container)) == NULL)
 		goto out_err;
+	if (req->header != NULL) {
+		mt->header = HX_strdup(req->header);
+		if (mt->header == NULL)
+			goto out_err;
+	}
 	if (S_ISBLK(sb.st_mode)) {
 		mt->loop_device  = NULL;
 		mt->lower_device = req->container;
@@ -404,7 +451,7 @@ EXPORT_SYMBOL int ehd_unload(struct ehd_mount_info *mt)
 }
 
 #ifndef HAVE_LIBCRYPTSETUP
-EXPORT_SYMBOL int ehd_is_luks(const char *device, bool blkdev)
+EXPORT_SYMBOL int ehd_is_luks(const char *device, const char *header, bool blkdev)
 {
 	return -EINVAL;
 }
