@@ -43,7 +43,7 @@ static unsigned int ehd_debug;
  */
 struct container_ctl {
 	unsigned long long size;
-	char *path, *fstype, *cipher, *hash, *user;
+	char *path, *fstype, *cipher, *hash, *user, *header;
 	unsigned int keybits, skip_random, uid;
 	bool blkdev;
 };
@@ -62,17 +62,17 @@ struct ehd_ctl {
 	bool interactive;
 };
 
-static bool ehd_check(const struct ehd_ctl *pg)
+static bool ehd_check(const char *path, const char *pretty_name,
+    unsigned int force_level, bool interactive)
 {
-	const struct container_ctl *cont = &pg->cont;
 	int ret, ask = 0;
 	struct stat sb;
 	bool exists;
 
-	printf("Creating a new container at %s\n", cont->path);
+	printf("Creating a new %s at %s\n", pretty_name, path);
 
 	/* First, check for hideous symlinks */
-	if (lstat(cont->path, &sb) < 0) {
+	if (lstat(path, &sb) < 0) {
 		if (errno == ENOENT) {
 			exists = false;
 		} else {
@@ -88,20 +88,20 @@ static bool ehd_check(const struct ehd_ctl *pg)
 
 		/* Get confirmation for overwriting files */
 		++ask;
-		ret = HX_readlink(&target, cont->path);
+		ret = HX_readlink(&target, path);
 		if (ret < 0) {
 			fprintf(stderr, "readlink %s: %s\n",
-			        cont->path, strerror(-ret));
+			        path, strerror(-ret));
 			return false;
 		}
 		printf("%s is a symlink and points to %s\n",
-		       cont->path, target);
+		       path, target);
 		HXmc_free(target);
 		/* Get extra confirmation */
 		++ask;
 
 		/* Now check for underlying device */
-		if (stat(cont->path, &sb) < 0) {
+		if (stat(path, &sb) < 0) {
 			if (errno != ENOENT) {
 				exists = false;
 			} else {
@@ -116,16 +116,15 @@ static bool ehd_check(const struct ehd_ctl *pg)
 		++ask;
 	}
 
-	if (pg->force_level < ask) {
+	if (force_level < ask) {
 		hxmc_t *tmp = NULL;
 
-		if (!pg->interactive) {
+		if (!interactive) {
 			printf("Not automatically overwriting file.\n");
 			return false;
 		}
 
-		printf("Do you really want to overwrite %s? (y/n)\n",
-		       cont->path);
+		printf("Do you really want to overwrite %s? (y/n)\n", path);
 		if (HX_getl(&tmp, stdin) == NULL)
 			return false;
 		if (HX_tolower(*tmp) != 'y')
@@ -136,6 +135,22 @@ static bool ehd_check(const struct ehd_ctl *pg)
 	return true;
 }
 
+static bool ehd_check_path(const struct ehd_ctl *pg)
+{
+	return ehd_check(pg->cont.path, "path", pg->force_level, pg->interactive);
+}
+
+static bool ehd_check_header(const struct ehd_ctl *pg)
+{
+	const struct container_ctl *cont = &pg->cont;
+
+	if (cont->header == NULL)
+		return true;
+	if (strcmp(cont->header, cont->path) == 0)
+		return true;
+	return ehd_check(cont->header, "header", pg->force_level, pg->interactive);
+}
+
 static bool ehd_xfer(int fd, size_t z)
 {
 #define BUFSIZE (65536 / sizeof(*buffer) * sizeof(*buffer))
@@ -232,6 +247,56 @@ static bool ehd_create_container(struct ehd_ctl *pg)
 	return ret;
 }
 
+static bool ehd_create_header(struct ehd_ctl *pg)
+{
+	struct container_ctl *cont = &pg->cont;
+	bool ret = false;
+	int fd = -1;
+
+	/* Header is not detached */
+	if (cont->header == NULL)
+		return true;
+	/* Container is formatted already */
+	if (strcmp(cont->path, cont->header) == 0)
+		return true;
+	fd = open(cont->header, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR);
+	if (fd < 0) {
+		fprintf(stderr, "open %s: %s\n", cont->header, strerror(errno));
+		return false;
+	}
+	if (cont->blkdev) {
+		ret = true;
+		goto out;
+	}
+
+	/*
+	 * /dev nodes should not be owned by user, even if it
+	 * is "their" voulme. Note that due to /dev being on tmpfs, ownership
+	 * is lost anyway after a reboot or device removal/add.
+	 */
+	if (fchown(fd, cont->uid, -1) < 0) {
+		fprintf(stderr, "fchown %s: %s\n", cont->header, strerror(errno));
+		ret = false;
+		goto out;
+	}
+
+	/*
+	 * Truncate on block devices does not make sense and would also return
+	 * EINVAL, so do not do it for block devices either.
+	 */
+	if (ftruncate(fd, 4096) < 0) {
+		fprintf(stderr, "ftruncate %s: %s\n", cont->header, strerror(errno));
+		ret = false;
+		goto out;
+	}
+
+	ret = true;
+ out:
+	if (fd >= 0)
+		close(fd);
+	return ret;
+}
+
 static bool ehd_mkfs(struct ehd_mount_request *rq,
     struct ehd_mount_info *mtinfo, void *priv)
 {
@@ -287,6 +352,7 @@ static int ehd_init_volume_luks(struct ehd_mount_request *rq,
 	struct crypt_params_luks1 format_params = {.hash = cont->hash};
 	struct crypt_device *cd = NULL;
 	const char *lower_dev = NULL;
+	const char *header = NULL;
 	int ret;
 
 	BUILD_BUG_ON(!__builtin_types_compatible_p(
@@ -297,10 +363,13 @@ static int ehd_init_volume_luks(struct ehd_mount_request *rq,
 	ret = ehd_mtinfo_get(mtinfo, EHD_MTINFO_LOWERDEV, &lower_dev);
 	if (ret <= 0 || lower_dev == NULL)
 		goto out;
-	ret = crypt_init(&cd, lower_dev);
+	ret = ehd_mtinfo_get(mtinfo, EHD_MTINFO_HEADER, &header);
+	if (ret <= 0)
+		goto out;
+	ret = crypt_init(&cd, header ?: lower_dev);
 	if (ret < 0) {
-		fprintf(stderr, "crypt_init: %s: %s\n",
-		        lower_dev, strerror(-ret));
+		fprintf(stderr, "crypt_init_data_device: %s, %s: %s\n",
+		        header, lower_dev, strerror(-ret));
 		goto out;
 	}
 	ret = crypt_format(cd, CRYPT_LUKS1, cipher, cipher_mode, NULL, NULL,
@@ -336,6 +405,9 @@ static bool ehd_init_volume(struct ehd_ctl *pg)
 	if (mount_request == NULL)
 		return -errno;
 	ret = ehd_mtreq_set(mount_request, EHD_MTREQ_CONTAINER, cont->path);
+	if (ret < 0)
+		goto out;
+	ret = ehd_mtreq_set(mount_request, EHD_MTREQ_HEADER, cont->header);
 	if (ret < 0)
 		goto out;
 	ret = ehd_mtreq_set(mount_request, EHD_MTREQ_KEY_SIZE,
@@ -378,10 +450,12 @@ static void ehd_final_printout(const struct ehd_ctl *pg)
 {
 	printf(
 		"-- (The important parts of) the new entry:\n\n"
-		"<volume fstype=\"crypt\" path=\"%s\" "
-		"mountpoint=\"REPLACEME\" />\n\n"
-		"-- Substitute paths by absolute ones.\n\n",
+		"<volume fstype=\"crypt\" path=\"%s\" ",
 		pg->cont.path);
+	if (pg->cont.header && strcmp(pg->cont.path, pg->cont.header) != 0)
+		printf("header=\"%s\" ", pg->cont.header);
+	printf("mountpoint=\"REPLACEME\" />\n\n"
+		"-- Substitute paths by absolute ones.\n\n");
 }
 
 /**
@@ -426,6 +500,10 @@ static bool ehd_fill_options_container(struct ehd_ctl *pg)
 		cont->path = HX_strdup(tmp);
 	}
 
+	ret = ehd_realpath(&cont->path);
+	if (ret < 0)
+		goto out;
+
 	if (stat(cont->path, &sb) < 0) {
 		if (errno != ENOENT) {
 			fprintf(stderr, "stat %s: %s\n",
@@ -442,6 +520,28 @@ static bool ehd_fill_options_container(struct ehd_ctl *pg)
 		}
 	}
 
+	if (cont->header == NULL && pg->interactive) {
+		printf("LUKS header path (optional): ");
+		fflush(stdout);
+		*tmp = '\0';
+		HX_getl(&tmp, stdin);
+		HX_chomp(tmp);
+		if (*tmp != '\0')
+			cont->header = HX_strdup(tmp);
+	}
+
+	if (cont->header != NULL) {
+		ret = ehd_realpath(&cont->header);
+		if (ret < 0)
+			goto out;
+	}
+
+	if (cont->header && stat(cont->header, &sb) < 0 && errno != ENOENT) {
+		fprintf(stderr, "stat %s: %s\n",
+		        cont->header, strerror(errno));
+		goto out;
+	}
+
 	if (cont->size == 0) {
 		unsigned int s;
 
@@ -515,6 +615,8 @@ static bool ehd_get_options(int *argc, const char ***argv, struct ehd_ctl *pg)
 		{.sh = 'h', .type = HXTYPE_STRING, .ptr = &cont->hash,
 		 .help = "Name of hash to be used for master keys (cryptsetup name)",
 		 .htyp = "NAME"},
+		{.sh = 'H', .type = HXTYPE_STRING, .ptr = &cont->header,
+		 .help = "Path of the new LUKS header file", .htyp = "FILE"},
 		{.sh = 'k', .type = HXTYPE_UINT, .ptr = &cont->keybits,
 		 .help = "Number of bits fscipher (-c) operates with",
 		 .htyp = "BITS"},
@@ -550,10 +652,14 @@ static int main2(int argc, const char **argv, struct ehd_ctl *pg)
 	if (!ehd_get_options(&argc, &argv, pg))
 		return EXIT_FAILURE;
 
-	if (!ehd_check(pg))
+	if (!ehd_check_path(pg))
 		return EXIT_FAILURE;
 	if (!ehd_create_container(pg))
 		return EXIT_FAILURE;
+	if (!ehd_check_header(pg))
+		return EXIT_FAILURE;
+	if (!ehd_create_header(pg))
+		return EXIT_FAILURE;
 	printf("NOTE: Use the end-user passphrase here. DO NOT feed it a hash "
 	       "of a passphrase or anything otherwise fancy. pmt-ehd(8), as "
 	       "well as cryptsetup(8)'s luksFormat will, by default, do the "
