@@ -645,6 +645,11 @@ int do_mount(const struct config *config, struct vol *vpt,
 			return 0;
 		}
 	}
+	if (password == NULL && !vpt->empty_pass) {
+		l0g("skipping volume %s because of empty password\n",
+		    vpt->volume);
+		return 0;
+	}
 
 	if (config->command[vpt->type]->items == 0) {
 		l0g("proper mount command not defined in "
