@@ -31,6 +31,7 @@
  * 			(file) /home/user.!@#)
  * @crypto_name:	preferred name of the crypto device (where possible)
  * @mountpoint:		where to put this
+ * @header:		LUKS header location (file or block device)
  * @extra_options:	options to pass down
  * @no_update:		do not update mtab
  * @loop_device:	loop device association, if any
@@ -42,20 +43,19 @@
  * @fsck:		true if fsck should be performed
  * @remount:		issue a remount
  * @allow_discards:	set block device to allow fs trim requests
+ * @no_read_workqueue: bypass internal workqueue and process read requests synchronously
+ * @no_write_workqueue: bypass internal workqueue and process write requests synchronously
  */
 struct mount_options {
-	hxmc_t *object, *container, *mountpoint;
+	hxmc_t *object, *container, *mountpoint, *header;
 	const char *fstype;
 	const char *crypto_name, *dmcrypt_cipher, *dmcrypt_hash;
 	const char *fsk_hash, *fsk_cipher, *fsk_file;
 	hxmc_t *fsk_password, *extra_opts, *crypto_device;
 	char *loop_device;
 	unsigned int no_update, readonly, trunc_keysize;
-	bool is_cont;
-	bool blkdev;
-	bool fsck;
-	bool remount;
-	bool allow_discards;
+	bool is_cont, blkdev, fsck, remount, allow_discards;
+	bool no_read_workqueue, no_write_workqueue;
 };
 
 /**
@@ -127,6 +127,8 @@ static void mtcr_parse_suboptions(const struct HXoptcb *cbi)
                 }
 		else if (strcmp(key, "keyfile") == 0)
 			mo->fsk_file = value;
+		else if (strcmp(key, "header") == 0)
+			mo->header = value;
 		else if (strcmp(key, "keysize") == 0)
 			mo->trunc_keysize = strtoul(value, NULL, 0) / CHAR_BIT;
 		else if (strcmp(key, "fsck") == 0)
@@ -150,6 +152,10 @@ static void mtcr_parse_suboptions(const struct HXoptcb *cbi)
 			mo->crypto_name = value;
 		} else if (strcmp(key, "allow_discard") == 0) {
 			mo->allow_discards = true;
+		} else if (strcmp(key, "no_read_workqueue") == 0) {
+			mo->no_read_workqueue = true;
+		} else if (strcmp(key, "no_write_workqueue") == 0) {
+			mo->no_write_workqueue = true;
 		} else {
 			/*
 			 * Above are the pam_mount-specific options that are
@@ -310,6 +316,22 @@ static bool mtcr_get_mount_options(int *argc, const char ***argv,
 		return false;
 	}
 
+	if (opt->header != NULL) {
+		ret = ehd_realpath(&opt->header);
+		if (ret < 0)
+			return false;
+
+		if (stat(opt->header, &sb) < 0) {
+			fprintf(stderr, "%s: stat %s: %s\n", **argv, opt->header,
+			        strerror(errno));
+			return false;
+		} else if (S_ISDIR(sb.st_mode)) {
+			fprintf(stderr, "%s: %s is a directory\n",
+			        **argv, opt->header);
+			return false;
+		}
+	}
+
 	kfpt = kfpt_selected(opt->fsk_cipher);
 	if (opt->fsk_file == NULL) {
 		if (opt->fsk_cipher != NULL)
@@ -330,7 +352,7 @@ static bool mtcr_get_mount_options(int *argc, const char ***argv,
 		}
 	}
 
-	ret = ehd_is_luks(opt->container, opt->blkdev);
+	ret = ehd_is_luks(opt->container, opt->header, opt->blkdev);
 	if (ret > 0) {
 		/* LUKS */
 		if (opt->dmcrypt_cipher != NULL)
@@ -454,6 +476,9 @@ static int mtcr_mount(struct mount_options *opt)
 	if (ret < 0)
 		goto out_r;
 	ret = ehd_mtreq_set(mount_request, EHD_MTREQ_CRYPTONAME, opt->crypto_name);
+	if (ret < 0)
+		goto out_r;
+	ret = ehd_mtreq_set(mount_request, EHD_MTREQ_HEADER, opt->header);
 	if (ret < 0)
 		goto out_r;
 	ret = ehd_mtreq_set(mount_request, EHD_MTREQ_FS_CIPHER, opt->dmcrypt_cipher);
@@ -469,6 +494,14 @@ static int mtcr_mount(struct mount_options *opt)
 	      opt->allow_discards);
 	if (ret < 0)
 		goto out_r;
+	ret = ehd_mtreq_set(mount_request, EHD_MTREQ_NO_READ_WORKQUEUE,
+	      opt->no_read_workqueue);
+	if (ret < 0)
+		goto out_r;
+	ret = ehd_mtreq_set(mount_request, EHD_MTREQ_NO_WRITE_WORKQUEUE,
+	      opt->no_write_workqueue);
+	if (ret < 0)
+		goto out_r;
 	/* Hack for CRYPT_PLAIN: default to 256 */
 	trunc_keysize = 256 / CHAR_BIT;
 	ret = ehd_mtreq_set(mount_request, EHD_MTREQ_TRUNC_KEYSIZE, trunc_keysize);
