@@ -1359,6 +1359,12 @@ static const char *rc_volume(xmlNode *node, struct config *config,
 		/* Figure out whether we want to act as user. */
 		vpt->noroot = strcmp(vpt->fstype, "fuse") == 0;
 
+	/* Empty password behavior. Default to trying empty passwords */
+	if ((tmp = xml_getprop(node, "empty_pass")) != NULL)
+		vpt->empty_pass = parse_bool_f(tmp);
+	else
+		vpt->empty_pass = true;
+
 	/* Source location */
 	if ((tmp = xml_getprop(node, "server")) != NULL) {
 		free(vpt->server);
